;;;;;;;;;;;;;;;;;;;;;;;
;;;                 ;;;
;;; Runtime Support ;;;
;;;                 ;;;
;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;; Top-level Scheme definitions given in this file are the preamble of
;;; all resulting Scheme object files generated from Scheme-compiling
;;; Infer source files in the Infer runtime system.
;;;
;;; Ideally, only a portion of these definitions are genuinely necessary
;;; for any specific Scheme (object) file though.
;;;
;;; This file is to be loaded on top-level Scheme.
;;;



;;;
;;; === HOST-LANGUAGE-DEPENDENT ===
;;;

;;; Define 'Standard Scheme' variables/mascros used in the rest of
;;; this file that are not supported by the host Scheme dialect


;;; Define non-'Standard Scheme' variables/macros used in the rest of
;;; this file that are not supported by the host Scheme dialect or not
;;; found elsewhere in this file

; The non-'Standard Scheme' variables involved are listed as below:
;   box? 



;;;
;;; Define *special* variables/macros used in the Scheme object code
;;; generated by the Infer runtime system ...
;;;

(define *apply* apply)
(define *car* car)
(define *cdr* cdr)
(define *cons* cons)
(define *false* #f)
(define *null?* null?)
(define *set-car!* set-car!)
(define *set-cdr!* set-cdr!)
(define *vector* vector)
(define *vector-ref* vector-ref)
(define *vector-set!* vector-set!)

(define (*forward* id)
  (lambda args
    (error id "UNDEFINED PROCEDURE")))

(define (*con?* con)
  (lambda (e)
    (eq? con (car e))))

(define (*con%* con)
  (lambda (e)
    (let ((con1 (car e)))
      (if (eq? con con1)
	  (cdr e)
	  (error '*con%*
	    "cannot project variant ~s given variant ~s" con con1)))))

(define (*set-con!* con)
  (lambda (e1 e2)
    (let ((con1 (car e1)))
      (if (eq? con con1)
	  (set-cdr! e1 e2)
	  (error '*set-con!*
	    "cannot mutate variant ~s given variant ~s" con con1)))))
			
(define (*cons%* e)
  (if (pair? e) e
    (error '*cons%* "cannot project variant cons given variant nil")))

(define (*opaque?* d)
  (not (or (number? d)
	   (boolean? d)
	   (string? d)
	   (symbol? d)
	   (char? d)
	   (vector? d)
	   (box? d)
	   (pair? d)
	   (null? d))))

(define (*d%* con)
  (let ((fn (case con
	      (num number?)
	      (char char?)
	      (sym symbol?)
	      (str string?)
	      (vec vector?)
	      (bool (lambda (x) #t))
	      (dbox box?)    
	      (pair pair?)
	      (else (lambda (x) #f)))))
    (lambda (e)
      (if (fn e) e
	(let ((con2 (cond
		      ((number? e) 'num)
		      ((char? e) 'char)
		      ((symbol? e) 'sym)
		      ((string? e) 'str)
		      ((vector? e) 'vec)
		      ((boolean? e) 'bool)
		      ((box? e) 'dbox)
		      ((pair? e) 'pair)
		      (else '??))))
	  (error '*d%* "cannot project variant ~s given variant ~s" con
	    con2))))))



;;;
;;; Give definitions to those Infer 'predefined' variables that do not
;;; have appropriate counterparts in the host Scheme environment ...
;;;

;;; See "runtime.is" for details.

(define unit '())

; (define nil nil)
; (define cons cons)
(define nil? null?)
(define cons? pair?)
(define cons% *cons%*)
(define cons->hd car)
(define cons->tl cdr)
;
(define empty '())
(define ?? (lambda (x) x))
(define num (lambda (x) x))
(define char (lambda (x) x))
(define sym (lambda (x) x))
(define str (lambda (x) x))
(define vec (lambda (x) x))
(define bool (lambda (x) x))
(define dbox (lambda (x) x))
(define pair cons)

(define empty? null?)
(define ??? *opaque?*)
(define num? number?)
; (define char? char?)
(define sym? symbol?)
(define str? string?)
(define vec? vector?)
(define bool? boolean?)
(define dbox? box?)    
; (define pair? pair?)

(define num% (*d%* 'num))
(define char% (*d%* 'char))
(define sym% (*d%* 'sym))
(define bool% (*d%* 'bool))
(define str% (*d%* 'str))
(define vec% (*d%* 'vec))
(define dbox% (*d%* 'dbox))    
(define pair% (*d%* 'pair))

(define pair->left car)
(define pair->right cdr)


;; loading the macro "local" 
;(load "local.ss")

;;; Implementation of local
;;; ML Style local declaration. Can be used anywhere DEFINE can be
;;; used. Does not support the (DEFINE (name args ...) body ...)
;;; syntax.
;;; Usage:

;;; decl::= (DEFINE name exp)
;;;       | (BEGIN decl ...)
;;;       | (LOCAL (decl ...) decl decl ...)

(define *do-local  
  (let ((set-difference
	  (lambda (s1 s2)
	    (let loop ([s s1] [ans '()])
	      (cond
		[(null? s) (reverse ans)]
		[(member (car s) s2) (loop (cdr s) ans)]
		[else (loop (cdr s) (cons (car s) ans))])))))
    (lambda (x)
      (syntax-case x (define)
	[(((define local-name local-exp) ...)
	  (define name exp)
	  ...)
	 (with-syntax
	   ([(name* ...)
	     (set-difference (syntax (name ...)) (syntax (local-name ...)))]
	    [(ignored-name) (generate-temporaries '(ignored))])
	   (syntax
	     (begin
	       (define name* #f)
	       ...
	       (define ignored-name ; gensymed name
		 (internal-let ((define local-name local-exp) ...)
		   (set! name exp)
		   ...)))))]
	[(local-decls decl ...)
	 (with-syntax
	   ([local-defns (*flatten (syntax (local-decls ())))]
	    [(defn ...) (*flatten (syntax ((decl ...) ())))])
	   (*do-local (syntax (local-defns defn ...))))]))))

(define *flatten
  (lambda (x)
    (syntax-case x (define begin local)
      [(() defns) (syntax defns)]
      [(((define name exp) more ...) (defn ...))
       (*flatten (syntax ((more ...) (defn ... (define name exp)))))]
      [(((begin decl ...) more ...) defns)
       (*flatten (syntax ((decl ... more ...) defns)))]
      [(((local local-decls decl ...) more ...) defns)
       (with-syntax
	 ([decl (*do-local (syntax (local-decls decl ...)))])
	 (*flatten (syntax ((decl more ...) defns))))])))

(define-syntax internal-let
  (letrec ((remove-duplicates
	     (lambda (ls)
	       (if (null? ls) ls
		 (if (member (car ls) (cdr ls))
		   (remove-duplicates (cdr ls))
		   (cons (car ls) (remove-duplicates (cdr ls))))))))
    (lambda (x)
      (syntax-case x (define)
	[(_ ((define name exp) ...) body ...)
	 (with-syntax
	   ([(name* ...) (remove-duplicates (syntax (name ...)))])
	   (syntax
	     (let ([name* #f] ...)
	       (set! name exp) ...
	       body ...)))]
	[(_ decls body ...)
	 (with-syntax
	   ([defns (*flatten (syntax (decls ())))])
	   (syntax (internal-let defns body ...)))]))))

(define-syntax local
  (lambda (x)
    (syntax-case x ()
      [(_ more ...) (*do-local (syntax (more ...)))])))




;;;
;;; Give definitions to those Infer 'preamble' variables that do not
;;; have appropriate counterparts in the host Scheme environment ...
;;;

;;; See the file "preamble.is" for a complete listing.

;;; Type declarations

(define absent '(absent))
(define (absent? o) (and (pair? o) (null? (cdr o)) (eq? (car o) 'absent)))
(define (present v) `(present . ,v))
(define (present? o) (and (pair? o) (eq? (car o) 'present)))
(define (present% o)
  (if (present? o) (cdr o)
    (error 'present% "cannot project variant present given variant absent")))


;;; Standard variables


;;; Other variables
     
(define (assoc? id ls)
  (let ((ans (assoc id (map vector->list ls))))
    (if ans (present (cadr ans))
      absent)))

(define (assq? id ls)
  (let ((ans (assq id (map vector->list ls))))
    (if ans (present (cadr ans))
      absent)))

(define (assv? id ls)
  (let ((ans (assv id (map vector->list ls))))
    (if ans (present (cadr ans))
      absent)))

(define dappend append)

(define (datum->list d)
  (if (list? d) d (error 'datum->list "~a is not a list.")))

(define (error sy . rest)    ; re-defined since printf is redefined (see later)
  (let ((st (if (null? rest) "" (car rest)))
	(datums (if (null? rest) nil (cdr rest))))
    (printf "~%ERROR in ~s" sy)
    (unless (string=? st "")
      (display ": ")
      (apply printf st datums)
      (newline))
    (newline)
    (reset)))

(define (fprintf port str . datums)  ; more expressive than the one in Chez
  (let ((len (string-length str)))
    (let f ((i 0) (ls datums))
      (unless (= i len)
        (let ((ch (string-ref str i)))
          (cond
            ((and (char=? ch #\~)
	       (< (+ i 1) len))
             (case (string-ref str
                     (+ i 1))
               ((#\s)
                (write (car ls) port)
                (f (+ i 2) (cdr ls)))
               ((#\a)
                (display (car ls) port)
                (f (+ i 2) (cdr ls)))
               ((#\p)
                (pretty-print (car ls)
                  port)
                (f (+ i 2) (cdr ls)))
               ((#\% #\n)
                (newline port)
                (f (+ i 2) ls))
               ((#\~)
                (write-char #\~ port)
                (f (+ i 2) ls))
               (else
                 (write-char #\~ port)
                 (f (+ i 1) ls))))
	    (else
	      (write-char ch)
	      (f (+ i 1) ls))))))))

(define left car)
(define (list->datum l) l)
(define make-forward *forward*)
(define member? member)
(define memq? memq)
(define memv? memv)

(define printf  ; Re-defined in Chez since fprintf is redefined (see above)
  (lambda (str . datums)
    (apply fprintf
      (current-output-port)
      str datums)))

(define quit abort)
(define right cdr)
(define (throw k) k)










